using System.Threading;
using System.Linq;
using System.Collections.Generic;
using UnityEngine;
using System.Collections.Concurrent;
using System;

[System.Serializable]
public struct Octave
{
    public float amplitude;
    public float frequency;
    public NoiseType type;
}

[System.Serializable]
public enum NoiseType
{
    MathfPerlin,
    MathematicsPerlin,
    Simplex,
    Celluar
}

public class TerrainGenerator : MonoBehaviour
{
    [Header("User settings")]
    public List<Octave> octaves;

    [Tooltip("The seed of the generation.")]
    public Vector2 crawlStartPosition;

    public bool generateNewSeed = true;

    public bool refresh;

    [Tooltip("How fine grain the generated terrain will be. " +
        "Larger values = more precise. Recommended value is 513.")]
    public int samples = 513;

    [Tooltip("How tall the terrain is. May be changed to allow for higher terrains.")]
    public float terrainHeight = 600;

    [Tooltip("What percentage of threads available should we use?")]
    public float threadsUsage = 0.7f;

    public int MaxThreads => Mathf.FloorToInt(Environment.ProcessorCount * threadsUsage);


    [Header("Autogenerated values")]
    [Tooltip("Which terrain grid are we on right now?")]
    public Vector2Int currentGridPosition;

    [Tooltip("Dictionary of terrains to reuse.")]
    public Dictionary<Vector2Int, Terrain> terrains = new();

    // [Tooltip("Dictionary of previously loaded terrains.")]
    // public Dictionary<Vector2Int, TerrainData> previouslyLoaded = new();

    [Tooltip("Contains the grid positions claimed by a generator thread.")]
    public HashSet<Vector2Int> claims = new();

    public ConcurrentQueue<TerrainGenThread> queuedThreads = new();

    public ConcurrentBag<TerrainGenThread> completedGenerators = new();


    [Header("Stats")]
    public int runningThreads = 0;

    private bool ReuseTerrain(Vector2Int gridPos, out Terrain terrain)
    {
        if (terrains.Count > 0)
        {
            var furthestTerrainPair = terrains.Aggregate((tp1, tp2) => 
                SquareDistance(gridPos, tp1.Key) > SquareDistance(gridPos, tp2.Key) ?
                tp1 : tp2);
            
            if (SquareDistance(furthestTerrainPair.Key, gridPos) > 1)
            {
                terrain = furthestTerrainPair.Value;
                terrains.Remove(furthestTerrainPair.Key);
                return true;
            }
        }

        // Either furthest terrain is visible or no terrains to reuse.
        terrain = null;
        return false;
    }

    /// <summary>
    /// Apply the terrain that was generated by genThread.
    /// </summary>
    /// <param name="genThread">The terrain gen thread that has finished 
    /// generating its terrain heights.</param>
    private void ApplyGeneratedTerrain(TerrainGenThread genThread)
    {
        genThread.TerrainData.SetHeights(0, 0, genThread.GetHeights());
        GameObject obj;

        if (ReuseTerrain(currentGridPosition, out Terrain reused))
        {
            // Reuse terrain
            obj = reused.gameObject;
            reused.terrainData = genThread.TerrainData;
            terrains[genThread.GridPosition] = reused;
            print($"Reusing {reused}");
        }
        else
        {
            // Create new terrain
            obj = Terrain.CreateTerrainGameObject(genThread.TerrainData);
            terrains[genThread.GridPosition] = obj.GetComponent<Terrain>();
        }

        // // Create new terrain
        // obj = Terrain.CreateTerrainGameObject(genThread.TerrainData);
        // terrains[genThread.GridPosition] = obj.GetComponent<Terrain>();
        
        obj.name = genThread.InitialCrawlPosition.ToString() + "Terrain";
        obj.layer = LayerMask.NameToLayer("Terrain");
        obj.transform.SetParent(transform, false);
        Vector3 gridWorldPosition = new Vector3(genThread.GridPosition.x, 0, genThread.GridPosition.y) *
            genThread.TerrainData.size.x;
        Vector3 centeringOffset = new Vector3(genThread.TerrainData.size.x, 0,
            genThread.TerrainData.size.z) / -2;
        obj.transform.localPosition = gridWorldPosition + centeringOffset;
    }

    private bool StakeClaims(Vector2Int gridPos)
    {
        // Don't claim if there already exists a terrain or if something
        // is already generating the terrain.
        if (claims.Contains(gridPos) || terrains.ContainsKey(gridPos))
        {
            return false;
        }
        else
        {
            claims.Add(gridPos);
            return true;
        }
    }

    private void GenerateThread(Vector2 crawlPos, Vector2Int gridPos,
         bool forceRefresh)
    {
        // No need to check for generated terrains. Should already been done.
        TerrainData data = Instantiate(Resources.Load<TerrainData>("Template"));
        TerrainGenThread tGenThread = new(this, data, crawlPos, gridPos);

        if (runningThreads >= MaxThreads)
        {
            // Too many threads.
            queuedThreads.Enqueue(tGenThread);
        }
        else
        {
            StartThread(tGenThread);
        }
    }

    private void StartThread(TerrainGenThread tGenThread)
    {
        runningThreads++;
        ThreadStart starter = new(tGenThread.Generate);
        Thread thread = new(starter);
        thread.Start();
    }

    public virtual void GenerateSurrounding(Vector2 initCrawlPos, bool forceRefresh = false)
    {
        List<Vector2Int> unclaimedGridPos = new();

        // Do claims first, then generate the threads, because I really don't
        // want to deal with concurrency issues.
        for (int i = -1; i <= 1; i++)
        {
            for (int j = 1; j >= -1; j--)
            {
                Vector2Int gridPos = new Vector2Int(i, j) + currentGridPosition;
                if (StakeClaims(gridPos))
                    unclaimedGridPos.Add(gridPos);
            }
        }

        foreach (var gridPos in unclaimedGridPos)
        {
            GenerateThread(initCrawlPos + gridPos, gridPos,
                forceRefresh);
        }
    }

    private void Update()
    {
        // Check for refresh.
        if (refresh)
        {
            refresh = false;

            foreach (Transform child in transform)
                Destroy(child.gameObject);

            GenerateSurrounding(crawlStartPosition, true);
        }

        // Check for queued threads.
        if (queuedThreads.TryDequeue(out TerrainGenThread queuedThread)
            && runningThreads < MaxThreads)
        {
            StartThread(queuedThread);
        }

        // Check for completed generators.
        if (completedGenerators.TryTake(out TerrainGenThread completedThread))
        {
            claims.Remove(completedThread.GridPosition);
            ApplyGeneratedTerrain(completedThread);
            runningThreads--;
        }
    }

    /// <summary>
    /// Gets the square distance between two points.
    /// If A is located at (0,0) and B is located at (5,-10),
    /// the distance would be 5+10=15.
    /// </summary>
    /// <param name="a">A position on the grid.</param>
    /// <param name="b">Another position.</param>
    /// <returns></returns>
    public int SquareDistance(Vector2Int a, Vector2Int b)
    {
        Vector2Int diff = b - a;

        return Mathf.Max(Mathf.Abs(diff.x), Mathf.Abs(diff.y));
    }
}
